#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""FAT and BPB parsing for files."""

import errno
import itertools

import math
import struct
import threading
import warnings

from contextlib import contextmanager
from io import BufferedReader, open

from pyfatfs import FAT_OEM_ENCODING
from pyfatfs.EightDotThree import EightDotThree
from pyfatfs.FATDirectoryEntry import FATDirectoryEntry, FATLongDirectoryEntry
from pyfatfs._exceptions import PyFATException, NotAFatEntryException


def _init_check(func):
    def _wrapper(*args, **kwargs):
        initialised = args[0].initialised

        if initialised is True:
            return func(*args, **kwargs)
        else:
            raise PyFATException("Class has not yet been fully initialised, "
                                 "please instantiate first.")

    return _wrapper


def _readonly_check(func):
    def _wrapper(*args, **kwargs):
        read_only = args[0].is_read_only

        if read_only is False:
            return func(*args, **kwargs)
        else:
            raise PyFATException("Filesystem has been opened read-only, not "
                                 "able to perform a write operation!")

    return _wrapper


class PyFat(object):
    """PyFAT base class, parses generic filesystem information."""

    #: Used as fat_type if unable to detect FAT type
    FAT_TYPE_UNKNOWN = 0
    #: Used as fat_type if FAT12 fs has been detected
    FAT_TYPE_FAT12 = 12
    #: Used as fat_type if FAT16 fs has been detected
    FAT_TYPE_FAT16 = 16
    #: Used as fat_type if FAT32 fs has been detected
    FAT_TYPE_FAT32 = 32

    #: Maps fat_type to BS_FilSysType from FS header information
    FS_TYPES = {FAT_TYPE_UNKNOWN: b"FAT     ",
                FAT_TYPE_FAT12: b"FAT12   ",
                FAT_TYPE_FAT16: b"FAT16   ",
                FAT_TYPE_FAT32: b"FAT32   "}

    #: Possible cluster values for FAT12 partitions
    FAT12_CLUSTER_VALUES = {'FREE_CLUSTER': 0x000,
                            'MIN_DATA_CLUSTER': 0x002,
                            'MAX_DATA_CLUSTER': 0xFEF,
                            'BAD_CLUSTER': 0xFF7,
                            'END_OF_CLUSTER_MIN': 0xFF8,
                            'END_OF_CLUSTER_MAX': 0xFFF}
    FAT12_SPECIAL_EOC = 0xFF0
    #: Possible cluster values for FAT16 partitions
    FAT16_CLUSTER_VALUES = {'FREE_CLUSTER': 0x0000,
                            'MIN_DATA_CLUSTER': 0x0002,
                            'MAX_DATA_CLUSTER': 0xFFEF,
                            'BAD_CLUSTER': 0xFFF7,
                            'END_OF_CLUSTER_MIN': 0xFFF8,
                            'END_OF_CLUSTER_MAX': 0xFFFF}
    #: Possible cluster values for FAT32 partitions
    FAT32_CLUSTER_VALUES = {'FREE_CLUSTER': 0x0000000,
                            'MIN_DATA_CLUSTER': 0x0000002,
                            'MAX_DATA_CLUSTER': 0x0FFFFFEF,
                            'BAD_CLUSTER': 0xFFFFFF7,
                            'END_OF_CLUSTER_MIN': 0xFFFFFF8,
                            'END_OF_CLUSTER_MAX': 0xFFFFFFF}
    #: Maps fat_type to possible cluster values
    FAT_CLUSTER_VALUES = {FAT_TYPE_FAT12: FAT12_CLUSTER_VALUES,
                          FAT_TYPE_FAT16: FAT16_CLUSTER_VALUES,
                          FAT_TYPE_FAT32: FAT32_CLUSTER_VALUES}

    #: BPB header layout in struct formatted string
    bpb_header_layout = "<3s8sHBHBHHBHHHLL"
    #: BPB header fields when extracted with bpb_header_layout
    bpb_header_vars = ["BS_jmpBoot", "BS_OEMName", "BPB_BytsPerSec",
                       "BPB_SecPerClus", "BPB_RsvdSecCnt", "BPB_NumFATS",
                       "BPB_RootEntCnt", "BPB_TotSec16", "BPB_Media",
                       "BPB_FATSz16", "BPB_SecPerTrk", "BPB_NumHeads",
                       "BPB_HiddSec", "BPB_TotSec32"]

    #: FAT12/16 header layout in struct formatted string
    fat12_header_layout = "<BBBL11s8s"
    #: FAT12/16 header fields when extracted with fat12_header_layout
    fat12_header_vars = ["BS_DrvNum", "BS_Reserved1", "BS_BootSig",
                         "BS_VolID", "BS_VolLab", "BS_FilSysType"]

    #: FAT32 header layout in struct formatted string
    fat32_header_layout = "<LHHLHH12sBBBL11s8s"
    #: FAT32 header fields when extracted with fat32_header_layout
    fat32_header_vars = ["BPB_FATSz32", "BPB_ExtFlags", "BPB_FSVer",
                         "BPB_RootClus", "BPB_FSInfo", "BPB_BkBootSec",
                         "BPB_Reserved", "BS_DrvNum", "BS_Reserved1",
                         "BS_BootSig", "BS_VolID", "BS_VolLab",
                         "BS_FilSysType"]

    #: FAT12/16 bit mask for clean shutdown bit
    FAT12_CLEAN_SHUTDOWN_BIT_MASK = 0x8000
    #: FAT12/16 bit mask for volume error bit
    FAT12_DRIVE_ERROR_BIT_MASK = 0x4000
    #: FAT32 bit mask for clean shutdown bit
    FAT32_CLEAN_SHUTDOWN_BIT_MASK = 0x0800000
    #: FAT32 bit mask for volume error bit
    FAT32_DRIVE_ERROR_BIT_MASK = 0x0400000

    def __init__(self,
                 encoding: str = 'ibm437',
                 offset: int = 0):
        """Set up PyFat class instance.

        :param encoding: Define encoding to use for filenames
        :param offset: Offset of the FAT partition in the given file
        :type encoding: str
        :type offset: int
        """
        self.__fp = None
        self.__fp_offset = offset
        self.bpb_header = None
        self.fat_header = None
        self.root_dir = None
        self.root_dir_sector = 0
        self.root_dir_sectors = 0
        self.bytes_per_cluster = 0
        self.first_data_sector = 0
        self.fat_type = self.FAT_TYPE_UNKNOWN
        self.fat = {}
        self.initialised = False
        self.fat_clusterchains = {}
        self.encoding = encoding
        self.is_read_only = True
        self.__lock = threading.Lock()

    def __set_fp(self, fp):
        if isinstance(self.__fp, BufferedReader):
            raise PyFATException("Cannot overwrite existing file handle, "
                                 "create new class instance of PyFAT.")
        self.__fp = fp

    def __seek(self, address: int):
        """Seek to given address with offset."""
        if self.__fp is None:
            raise PyFATException("Cannot seek without a file handle!",
                                 errno=errno.ENXIO)
        self.__fp.seek(address + self.__fp_offset)

    @_init_check
    def read_cluster_contents(self, cluster: int) -> bytes:
        """Read contents of given cluster.

        :param cluster: Cluster number to read contents from
        :returns: Contents of cluster as `bytes`
        """
        sz = self.bytes_per_cluster
        cluster_address = self.get_data_cluster_address(cluster)
        with self.__lock:
            self.__seek(cluster_address)
            return self.__fp.read(sz)

    def open(self, filename: str, read_only: bool = False):
        """Open filesystem for usage with PyFat.

        :param filename: `str`: Name of file to open for usage with PyFat.
        :param read_only: `bool`: Force read-only mode of filesystem.
        """
        self.is_read_only = read_only
        if read_only is True:
            mode = 'rb'
        else:
            mode = 'rb+'

        try:
            self.__set_fp(open(filename, mode=mode))
        except OSError as ex:
            raise PyFATException(f"Cannot open given file \'{filename}\'.",
                                 errno=ex.errno)

        # Parse BPB & FAT headers of given file
        self.parse_header()

        # Parse FAT
        self._parse_fat()

        # Parse root directory
        # TODO: Inefficient to always recursively parse the root dir.
        #       It would make sense to parse it on demand instead.
        self.parse_root_dir()

    @_init_check
    def get_fs_location(self):
        """Retrieve path of opened filesystem."""
        return self.__fp.name

    @_init_check
    def _get_total_sectors(self):
        """Get total number of sectors for all FAT sizes."""
        if self.bpb_header["BPB_TotSec16"] != 0:
            return self.bpb_header["BPB_TotSec16"]

        return self.bpb_header["BPB_TotSec32"]

    def _get_fat_size_count(self):
        if self.bpb_header["BPB_FATSz16"] != 0:
            return self.bpb_header["BPB_FATSz16"]

        # Only possible with FAT32
        self.__parse_fat32_header()
        try:
            return self.fat_header["BPB_FATSz32"]
        except KeyError:
            raise PyFATException("Invalid FAT size of 0 detected in header, "
                                 "cannot continue")

    @_init_check
    def _parse_fat(self):
        """Parse information in FAT."""
        # Read all FATs
        fat_size = self.bpb_header["BPB_BytsPerSec"]
        fat_size *= self._get_fat_size_count()

        # Seek FAT entries
        first_fat_bytes = self.bpb_header["BPB_RsvdSecCnt"]
        first_fat_bytes *= self.bpb_header["BPB_BytsPerSec"]
        fats = []
        for i in range(self.bpb_header["BPB_NumFATS"]):
            with self.__lock:
                self.__seek(first_fat_bytes + (i * fat_size))
                fats += [self.__fp.read(fat_size)]

        if len(fats) < 1:
            raise PyFATException("Invalid number of FATs configured, "
                                 "cannot continue")
        elif len(set(fats)) > 1:
            warnings.warn("One or more FATs differ, filesystem most "
                          "likely corrupted. Using first FAT.")

        # Parse first FAT
        self.bytes_per_cluster = self.bpb_header["BPB_BytsPerSec"] * \
            self.bpb_header["BPB_SecPerClus"]

        if len(fats[0]) != self.bpb_header["BPB_BytsPerSec"] * \
                self._get_fat_size_count():
            raise PyFATException("Invalid length of FAT")

        # FAT12: 12 bits (1.5 bytes) per FAT entry
        # FAT16: 16 bits (2 bytes) per FAT entry
        # FAT32: 32 bits (4 bytes) per FAT entry
        fat_entry_size = self.fat_type / 8
        total_entries = int(fat_size // fat_entry_size)
        self.fat = [None] * total_entries

        curr = 0
        cluster = 0
        incr = self.fat_type / 8
        while curr < fat_size:
            offset = curr + incr

            if self.fat_type == self.FAT_TYPE_FAT12:
                fat_nibble = fats[0][int(curr):math.ceil(offset)]
                self.fat[cluster] = struct.unpack("<H", fat_nibble)[0]

                if cluster % 2 == 0:
                    # Even: Keep low 12-bits of word
                    self.fat[cluster] &= 0x0FFF
                else:
                    # Odd: Keep high 12-bits of word
                    self.fat[cluster] >>= 4

                if math.ceil(offset) == (fat_size - 1):
                    # Sector boundary case for FAT12
                    del self.fat[-1]
                    break

            elif self.fat_type == self.FAT_TYPE_FAT16:
                self.fat[cluster] = struct.unpack("<H",
                                                  fats[0][int(curr):
                                                          int(offset)])[0]
            elif self.fat_type == self.FAT_TYPE_FAT32:
                self.fat[cluster] = struct.unpack("<L",
                                                  fats[0][int(curr):
                                                          int(offset)])[0]
                # Ignore first four bits, FAT32 clusters are
                # actually just 28bits long
                self.fat[cluster] &= 0x0FFFFFFF
            else:
                raise PyFATException("Unknown FAT type, cannot continue")

            curr += incr
            cluster += 1

        if None in self.fat:
            raise AssertionError("Unknown error during FAT parsing, please "
                                 "report this error.")

    @_init_check
    def byte_repr(self) -> bytearray:
        """Represent current state of FAT as bytes.

        :returns: `bytes` representation of FAT.
        """
        b = bytearray(b'')
        if self.fat_type == self.FAT_TYPE_FAT12:
            raise NotImplementedError("FAT12 write support currently "
                                      "not implemented!")
        elif self.fat_type == self.FAT_TYPE_FAT16:
            fmt = "<H"
        else:
            # FAT32
            fmt = "<L"

        for c in self.fat:
            b += struct.pack(fmt, c)
        return b

    @_init_check
    @_readonly_check
    def _write_data_to_address(self, data: bytes,
                               address: int):
        """Write given data directly to the filesystem.

        Directly writes to the filesystem without any consistency check.
        **Use with caution**

        :param data: `bytes`: Data to write to address
        :param address: `int`: Offset to write data to.
        """
        with self.__lock:
            self.__seek(address)
            self.__fp.write(data)

    @_init_check
    @_readonly_check
    def free_cluster_chain(self, cluster: int):
        """Mark a cluster(chain) as free in FAT.

        :param cluster: `int`: Cluster to mark as free
        """
        _freeclus = self.FAT_CLUSTER_VALUES[self.fat_type]['FREE_CLUSTER']
        with self.__lock:
            tmp_fat = self.fat.copy()
            for cl in self.get_cluster_chain(cluster):
                tmp_fat[cl] = _freeclus
            self.fat = tmp_fat

    @_init_check
    @_readonly_check
    def write_data_to_cluster(self, data: bytes,
                              cluster: int,
                              extend_cluster: bool = True,
                              erase: bool = False) -> None:
        """Write given data to cluster.

        Extends cluster chain if needed.

        :param data: `bytes`: Data to write to cluster
        :param cluster: `int`: Cluster to write data to.
        :param extend_cluster: `bool`: Automatically extend cluster chain
                               if not enough space is available.
        :param erase: `bool`: Erase cluster contents before writing.
                      This is useful when writing `FATDirectoryEntry` data.
        """
        data_sz = len(data)
        cluster_sz = 0
        last_cluster = None
        for c in self.get_cluster_chain(cluster):
            cluster_sz += self.bytes_per_cluster
            last_cluster = c

        if data_sz > cluster_sz:
            if extend_cluster is False:
                raise PyFATException("Cannot write data to cluster, "
                                     "not enough space available.",
                                     errno=errno.ENOSPC)

            new_chain = self.allocate_bytes(data_sz - cluster_sz,
                                            erase=erase)[0]
            self.fat[last_cluster] = new_chain

        # Fill rest of data with zeroes if erase is set to True
        if erase:
            new_sz = math.ceil(max(1, data_sz // self.bytes_per_cluster))
            new_sz *= self.bytes_per_cluster
            data += b'\0' * (new_sz - data_sz)

        # Write actual data
        bytes_written = 0
        for c in self.get_cluster_chain(cluster):
            b = self.get_data_cluster_address(c)
            t = bytes_written
            bytes_written += self.bytes_per_cluster
            self._write_data_to_address(data[t:bytes_written], b)

    @_init_check
    @_readonly_check
    def flush_fat(self) -> None:
        """Flush FAT(s) to disk."""
        fat_size = self.bpb_header["BPB_BytsPerSec"]
        fat_size *= self._get_fat_size_count()

        first_fat_bytes = self.bpb_header["BPB_RsvdSecCnt"]
        first_fat_bytes *= self.bpb_header["BPB_BytsPerSec"]

        for i in range(self.bpb_header["BPB_NumFATS"]):
            with self.__lock:
                self.__seek(first_fat_bytes + (i * fat_size))
                self.__fp.write(self.byte_repr())

    def calc_num_clusters(self, size: int = 0) -> int:
        """Calculate the number of required clusters.

        :param size: `int`: required bytes to allocate
        :returns: Number of required clusters
        """
        num_clusters = size / self.bytes_per_cluster
        num_clusters = math.ceil(num_clusters)

        return num_clusters

    @_init_check
    @_readonly_check
    def allocate_bytes(self, size: int, erase: bool = False) -> list:
        """Try to allocate a cluster (-chain) in FAT for `size` bytes.

        :param size: `int`: Size in bytes to try to allocate.
        :param erase: `bool`: If set to true, the newly allocated
                              space is zeroed-out for clean allocation.
        :returns: List of newly-allocated clusters.
        """
        free_clus = self.FAT_CLUSTER_VALUES[self.fat_type]["FREE_CLUSTER"]
        min_clus = self.FAT_CLUSTER_VALUES[self.fat_type]["MIN_DATA_CLUSTER"]
        max_clus = self.FAT_CLUSTER_VALUES[self.fat_type]["MAX_DATA_CLUSTER"]
        num_clusters = self.calc_num_clusters(size)

        # Fill list of found free clusters
        free_clusters = []
        for i, _ in enumerate(self.fat):
            if min_clus > i > max_clus:
                # Ignore out of bound entries
                continue

            if num_clusters == len(free_clusters):
                # Allocated enough clusters!
                break

            if self.fat[i] == free_clus:
                if i == self.FAT_CLUSTER_VALUES[self.fat_type]["BAD_CLUSTER"]:
                    # Do not allocate a BAD_CLUSTER
                    continue

                if self.fat_type == self.FAT_TYPE_FAT12 and \
                        i == self.FAT12_SPECIAL_EOC:
                    # Do not allocate special EOC marker on FAT12
                    continue

                free_clusters += [i]
        else:
            free_space = len(free_clusters) * self.bytes_per_cluster
            raise PyFATException(f"Not enough free space to allocate "
                                 f"{size} bytes ({free_space} bytes free)",
                                 errno=errno.ENOSPC)

        # Allocate cluster chain in FAT
        eoc_max = self.FAT_CLUSTER_VALUES[self.fat_type]["END_OF_CLUSTER_MAX"]
        for i, _ in enumerate(free_clusters):
            try:
                self.fat[free_clusters[i]] = free_clusters[i+1]
            except IndexError:
                self.fat[free_clusters[i]] = eoc_max

            if erase is True:
                with self.__lock:
                    self.__seek(self.get_data_cluster_address(i))
                    self.__fp.write(b'\0' * self.bytes_per_cluster)

        return free_clusters

    @_init_check
    @_readonly_check
    def update_directory_entry(self, dir_entry: FATDirectoryEntry) -> None:
        """Update directory entry on disk.

        Special handling is required, since the root directory
        on FAT12/16 is on a fixed location on disk.

        :param dir_entry: `FATDirectoryEntry`: Directory to write to disk
        """
        is_root_dir = False
        extend_cluster_chain = True
        if self.root_dir == dir_entry:
            if self.fat_type != self.FAT_TYPE_FAT32:
                # FAT12/16 doesn't have a root directory cluster,
                # which cannot be enhanced
                extend_cluster_chain = False
            is_root_dir = True

        # Gather all directory entries
        dir_entries = b''
        d, f, s = dir_entry.get_entries()
        for d in list(itertools.chain(d, f, s)):
            dir_entries += d.byte_repr()

        # Write content
        if not is_root_dir or self.fat_type == self.FAT_TYPE_FAT32:
            # FAT32 and non-root dir entries can be handled normally
            self.write_data_to_cluster(dir_entries,
                                       dir_entry.get_cluster(),
                                       extend_cluster=extend_cluster_chain,
                                       erase=True)
        else:
            # FAT12/16 does not have a root directory cluster
            root_dir_addr = self.root_dir_sector * \
                self.bpb_header["BPB_BytsPerSec"]
            root_dir_sz = self.root_dir_sectors * \
                self.bpb_header["BPB_BytsPerSec"]

            if len(dir_entries) > root_dir_sz:
                raise PyFATException("Cannot create directory, maximum number "
                                     "of root directory entries exhausted!",
                                     errno=errno.ENOSPC)

            # Overwrite empty space as well
            dir_entries += b'\0' * (root_dir_sz - len(dir_entries))
            self._write_data_to_address(dir_entries, root_dir_addr)

    def _fat12_parse_root_dir(self):
        """Parse FAT12/16 root dir entries.

        FAT12/16 has a fixed location of root directory entries
        and is therefore size limited (BPB_RootEntCnt).
        """
        root_dir_byte = self.root_dir_sector * \
            self.bpb_header["BPB_BytsPerSec"]
        self.root_dir.set_cluster(self.root_dir_sector //
                                  self.bpb_header["BPB_SecPerClus"])
        max_bytes = self.bpb_header["BPB_RootEntCnt"] * \
            FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE

        # Parse all directory entries in root directory
        subdirs, _ = self.parse_dir_entries_in_address(root_dir_byte,
                                                       root_dir_byte +
                                                       max_bytes)
        for dir_entry in subdirs:
            self.root_dir.add_subdirectory(dir_entry)

    def _fat32_parse_root_dir(self):
        """Parse FAT32 root dir entries.

        FAT32 actually has its root directory entries distributed
        across a cluster chain that we need to follow
        """
        root_cluster = self.fat_header["BPB_RootClus"]
        self.root_dir.set_cluster(root_cluster)

        # Follow root directory cluster chain
        for dir_entry in self.parse_dir_entries_in_cluster_chain(root_cluster):
            self.root_dir.add_subdirectory(dir_entry)

    def parse_root_dir(self):
        """Parse root directory entry."""
        root_dir_sfn = EightDotThree()
        root_dir_sfn.set_str_name("")
        dir_attr = FATDirectoryEntry.ATTR_DIRECTORY
        self.root_dir = FATDirectoryEntry(DIR_Name=root_dir_sfn,
                                          DIR_Attr=dir_attr,
                                          DIR_NTRes=0,
                                          DIR_CrtTimeTenth=0,
                                          DIR_CrtTime=0,
                                          DIR_CrtDate=0,
                                          DIR_LstAccessDate=0,
                                          DIR_FstClusHI=0,
                                          DIR_WrtTime=0,
                                          DIR_WrtDate=0,
                                          DIR_FstClusLO=0,
                                          DIR_FileSize=0,
                                          encoding=self.encoding)

        if self.fat_type in [self.FAT_TYPE_FAT12, self.FAT_TYPE_FAT16]:
            self._fat12_parse_root_dir()
        else:
            self._fat32_parse_root_dir()

    def parse_lfn_entry(self,
                        lfn_entry: FATLongDirectoryEntry = None,
                        address: int = 0):
        """Parse LFN entry at given address."""
        dir_hdr_sz = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE

        with self.__lock:
            self.__seek(address)
            lfn_dir_data = self.__fp.read(dir_hdr_sz)

        lfn_hdr_layout = FATLongDirectoryEntry.FAT_LONG_DIRECTORY_LAYOUT
        lfn_dir_hdr = struct.unpack(lfn_hdr_layout, lfn_dir_data)
        lfn_dir_hdr = dict(zip(FATLongDirectoryEntry.FAT_LONG_DIRECTORY_VARS,
                               lfn_dir_hdr))

        lfn_entry.add_lfn_entry(**lfn_dir_hdr)

    def __parse_dir_entry(self, address):
        """Parse directory entry at given address."""
        with self.__lock:
            self.__seek(address)
            dir_hdr_size = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE
            dir_data = self.__fp.read(dir_hdr_size)

        dir_hdr = struct.unpack(FATDirectoryEntry.FAT_DIRECTORY_LAYOUT,
                                dir_data)
        dir_hdr = dict(zip(FATDirectoryEntry.FAT_DIRECTORY_VARS, dir_hdr))
        return dir_hdr

    def parse_dir_entries_in_address(self,
                                     address: int = 0,
                                     max_address: int = 0,
                                     tmp_lfn_entry: FATLongDirectoryEntry =
                                     FATLongDirectoryEntry()):
        """Parse directory entries in address range."""
        dir_hdr_size = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE

        if max_address == 0:
            max_address = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE

        dir_entries = []

        for hdr_addr in range(address, max_address, dir_hdr_size):
            # Parse each entry
            dir_hdr = self.__parse_dir_entry(hdr_addr)
            dir_sn = EightDotThree(encoding=self.encoding)
            try:
                dir_sn.set_byte_name(dir_hdr["DIR_Name"])
            except NotAFatEntryException as ex:
                # Not a directory of any kind, invalidate temporary LFN entries
                tmp_lfn_entry = FATLongDirectoryEntry()
                if ex.free_type == ex.FREE_ENTRY:
                    # Empty directory entry,
                    continue
                elif ex.free_type == ex.LAST_ENTRY:
                    # Last directory entry, do not parse any further
                    break
            else:
                dir_hdr["DIR_Name"] = dir_sn

            # Long File Names
            if FATLongDirectoryEntry.is_lfn_entry(dir_hdr["DIR_Name"],
                                                  dir_hdr["DIR_Attr"]):
                self.parse_lfn_entry(tmp_lfn_entry, hdr_addr)
                continue

            # Normal directory entries
            if not tmp_lfn_entry.is_lfn_entry_complete():
                # Ignore incomplete LFN entries altogether
                tmp_lfn_entry = None

            dir_entry = FATDirectoryEntry(encoding=self.encoding,
                                          lfn_entry=tmp_lfn_entry,
                                          **dir_hdr)
            dir_entries += [dir_entry]

            if dir_entry.is_directory() and not dir_entry.is_special():
                # Iterate all subdirectories except for dot and dotdot
                cluster = dir_entry.get_cluster()
                subdirs = self.parse_dir_entries_in_cluster_chain(cluster)
                for d in subdirs:
                    dir_entry.add_subdirectory(d)

            # Reset temporary LFN entry
            tmp_lfn_entry = FATLongDirectoryEntry()

        return dir_entries, tmp_lfn_entry

    def parse_dir_entries_in_cluster_chain(self, cluster):
        """Parse directory entries while following given cluster chain."""
        dir_entries = []
        tmp_lfn_entry = FATLongDirectoryEntry()
        max_bytes = (self.bpb_header["BPB_SecPerClus"] *
                     self.bpb_header["BPB_BytsPerSec"])
        for c in self.get_cluster_chain(cluster):
            # Parse all directory entries in chain
            b = self.get_data_cluster_address(c)
            ret = self.parse_dir_entries_in_address(b, b+max_bytes,
                                                    tmp_lfn_entry)
            tmp_dir_entries, tmp_lfn_entry = ret
            dir_entries += tmp_dir_entries

        return dir_entries

    def get_data_cluster_address(self, cluster: int) -> int:
        """Get offset of given cluster in bytes.

        :param cluster: Cluster number as `int`
        :returns: Bytes address location of cluster
        """
        # First two cluster entries are reserved
        sector = (cluster - 2) * self.bpb_header["BPB_SecPerClus"] + \
            self.first_data_sector
        return sector * self.bpb_header["BPB_BytsPerSec"]

    @_init_check
    def get_cluster_chain(self, first_cluster):
        """Follow a cluster chain beginning with the first cluster address."""
        cluster_vals = self.FAT_CLUSTER_VALUES[self.fat_type]
        min_data_cluster = cluster_vals["MIN_DATA_CLUSTER"]
        max_data_cluster = cluster_vals["MAX_DATA_CLUSTER"]
        eoc_min = cluster_vals["END_OF_CLUSTER_MIN"]
        eoc_max = cluster_vals["END_OF_CLUSTER_MAX"]

        i = first_cluster
        while i <= len(self.fat):
            if min_data_cluster <= self.fat[i] <= max_data_cluster:
                # Normal data cluster, follow chain
                yield i
            elif self.fat_type == self.FAT_TYPE_FAT12 and \
                    self.fat[i] == self.FAT12_SPECIAL_EOC:
                # Special EOC
                yield i
                return
            elif eoc_min <= self.fat[i] <= eoc_max:
                # End of cluster, end chain
                yield i
                return
            elif self.fat[i] == cluster_vals["BAD_CLUSTER"]:
                # Bad cluster, cannot follow chain, file broken!
                raise PyFATException("Bad cluster found in FAT cluster "
                                     "chain, cannot access file")
            elif self.fat[i] == cluster_vals["FREE_CLUSTER"]:
                # FREE_CLUSTER mark when following a chain is treated an error
                raise PyFATException("FREE_CLUSTER mark found in FAT cluster "
                                     "chain, cannot access file")
            else:
                raise PyFATException("Invalid or unknown FAT cluster "
                                     "entry found with value "
                                     "\'{}\'".format(hex(self.fat[i])))

            i = self.fat[i]

    @_init_check
    def close(self):
        """Close session and free up all handles."""
        if not self.is_read_only:
            self.flush_fat()

        self.__fp.close()
        self.initialised = False

    def __del__(self):
        """Try to close open handles."""
        try:
            self.close()
        except PyFATException:
            pass

    def __determine_fat_type(self):
        """Determine FAT size.

        An internal method to determine whether this volume is FAT12,
        FAT16 or FAT32.

        returns: `str`: Any of PyFat.FAT_TYPE_FAT12, PyFat.FAT_TYPE_FAT16
                 or PyFat.FAT_TYPE_FAT32
        """
        if self.bpb_header["BPB_TotSec16"] != 0:
            total_sectors = self.bpb_header["BPB_TotSec16"]
        else:
            total_sectors = self.bpb_header["BPB_TotSec32"]

        rsvd_sectors = self.bpb_header["BPB_RsvdSecCnt"]
        fat_sz = self.bpb_header["BPB_NumFATS"] * self._get_fat_size_count()
        root_dir_sectors = self.root_dir_sectors
        data_sec = total_sectors - (rsvd_sectors + fat_sz + root_dir_sectors)
        count_of_clusters = data_sec // self.bpb_header["BPB_SecPerClus"]

        if count_of_clusters < 4085:
            fat_type = self.FAT_TYPE_FAT12
        elif count_of_clusters < 65525:
            fat_type = self.FAT_TYPE_FAT16
        else:
            fat_type = self.FAT_TYPE_FAT32

        return fat_type

    def __parse_fat12_header(self):
        """Parse FAT12/16 header."""
        with self.__lock:
            self.__seek(0)
            boot_sector = self.__fp.read(512)

        header = struct.unpack(self.fat12_header_layout,
                               boot_sector[36:][:26])
        self.fat_header = dict(zip(self.fat12_header_vars, header))

    def __parse_fat32_header(self):
        """Parse FAT32 header."""
        with self.__lock:
            self.__seek(0)
            boot_sector = self.__fp.read(512)

        header = struct.unpack(self.fat32_header_layout,
                               boot_sector[36:][:54])
        self.fat_header = dict(zip(self.fat32_header_vars, header))

    def parse_header(self):
        """Parse BPB & FAT headers in opened file."""
        with self.__lock:
            self.__seek(0)
            boot_sector = self.__fp.read(512)

        header = struct.unpack(self.bpb_header_layout, boot_sector[:36])
        self.bpb_header = dict(zip(self.bpb_header_vars, header))

        # Verify BPB headers
        self.__verify_bpb_header()

        # Calculate root directory sectors and starting point of root directory
        root_entries = self.bpb_header["BPB_RootEntCnt"]
        hdr_size = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE
        bytes_per_sec = self.bpb_header["BPB_BytsPerSec"]
        rsvd_secs = self.bpb_header["BPB_RsvdSecCnt"]
        num_fats = self.bpb_header["BPB_NumFATS"]

        self.root_dir_sectors = ((root_entries * hdr_size) +
                                 (bytes_per_sec - 1)) // bytes_per_sec
        self.root_dir_sector = rsvd_secs + (self._get_fat_size_count() *
                                            num_fats)

        # Calculate first data sector
        self.first_data_sector = rsvd_secs + (num_fats *
                                              self._get_fat_size_count()) + \
            self.root_dir_sectors

        # Determine FAT type
        self.fat_type = self.__determine_fat_type()

        # Parse FAT type specific header
        if self.fat_type in [self.FAT_TYPE_FAT12, self.FAT_TYPE_FAT16]:
            self.__parse_fat12_header()

            self.__verify_fat12_header()
        elif self.fat_type == self.FAT_TYPE_FAT32:
            # FAT32, probably - probe for it
            if self.bpb_header["BPB_FATSz16"] != 0:
                raise PyFATException(f"Invalid BPB_FATSz16 value of "
                                     f"'{self.bpb_header['BPB_FATSz16']}', "
                                     f"filesystem corrupt?",
                                     errno=errno.EINVAL)

            self.__parse_fat32_header()

            # TODO: Verify FAT32 header
        else:
            raise PyFATException("Unknown FAT filesystem type, "
                                 "filesystem corrupt?", errno=errno.EINVAL)

        # Check signature
        with self.__lock:
            self.__seek(510)
            signature = struct.unpack("<H", self.__fp.read(2))[0]

        if signature != 0xAA55:
            raise PyFATException(f"Invalid signature: \'{hex(signature)}\'.")

        # Initialisation finished
        self.initialised = True

    def __verify_fat12_header(self):
        """Verify FAT12/16 header for correctness."""
        # Extended boot signature had two different versions:
        # 0x00: Pre-DOS 4.0 BPB (no EBPB present)
        # 0x28: Only BS_VolID is present
        # 0x29: Full FAT12 bootsector present
        bs_bootsig = self.fat_header["BS_BootSig"]
        if bs_bootsig == 0x29:
            bs_filsystype = self.fat_header["BS_FilSysType"]
            if bs_filsystype not in (self.FS_TYPES[self.FAT_TYPE_UNKNOWN],
                                     self.FS_TYPES[self.fat_type]):
                print(f"BS_FilSysType contains {bs_filsystype}, unexpected "
                      f"for FAT{self.fat_type}")
        elif bs_bootsig not in (0x28, 0x00, 0x2a):
            raise PyFATException("Invalid FAT extended boot signature value "
                                 "encountered: '{}'".format(hex(bs_bootsig)))

    def __verify_bpb_header(self):
        """Verify BPB header for correctness."""
        if self.bpb_header["BS_jmpBoot"][0] == 0xEB:
            if self.bpb_header["BS_jmpBoot"][2] != 0x90:
                raise PyFATException("Boot code must end with 0x90")
        elif self.bpb_header["BS_jmpBoot"][0] == 0xE9:
            pass
        else:
            raise PyFATException("Boot code must start with 0xEB or "
                                 "0xE9. Is this a FAT partition?")

        #: 512,1024,2048,4096: As per fatgen103.doc
        byts_per_sec_range = [2**x for x in range(9, 13)]
        if self.bpb_header["BPB_BytsPerSec"] not in byts_per_sec_range:
            raise PyFATException(f"Expected one of {byts_per_sec_range} "
                                 f"bytes per sector, got: "
                                 f"\'{self.bpb_header['BPB_BytsPerSec']}\'.")

        #: 1,2,4,8,16,32,64,128: As per fatgen103.doc
        sec_per_clus_range = [2**x for x in range(8)]
        if self.bpb_header["BPB_SecPerClus"] not in sec_per_clus_range:
            raise PyFATException(f"Expected one of {sec_per_clus_range} "
                                 f"sectors per cluster, got: "
                                 f"\'{self.bpb_header['BPB_SecPerClus']}\'.")

        bytes_per_cluster = self.bpb_header["BPB_BytsPerSec"]
        bytes_per_cluster *= self.bpb_header["BPB_SecPerClus"]
        if bytes_per_cluster > 32768:
            warnings.warn("Bytes per cluster should not be more than 32K, "
                          "but got: {}K. Trying to continue "
                          "anyway.".format(bytes_per_cluster // 1024), Warning)

        if self.bpb_header["BPB_RsvdSecCnt"] == 0:
            raise PyFATException("Number of reserved sectors must not be 0")

        if self.bpb_header["BPB_Media"] not in [0xf0, 0xf8, 0xf9, 0xfa, 0xfb,
                                                0xfc, 0xfd, 0xfe, 0xff]:
            raise PyFATException("Invalid media type")

        if self.bpb_header["BPB_NumFATS"] < 1:
            raise PyFATException("At least one FAT expected, None found.")

        root_entry_count = self.bpb_header["BPB_RootEntCnt"] * 32
        root_entry_count %= self.bpb_header["BPB_BytsPerSec"]
        if self.bpb_header["BPB_RootEntCnt"] != 0 and root_entry_count != 0:
            raise PyFATException("Root entry count does not cleanly align with"
                                 " bytes per sector!")

        if self.bpb_header["BPB_TotSec16"] == 0 and \
                self.bpb_header["BPB_TotSec32"] == 0:
            raise PyFATException("16-Bit and 32-Bit total sector count "
                                 "value empty.")

    @staticmethod
    @contextmanager
    def open_fs(filename: str, offset: int = 0,
                encoding=FAT_OEM_ENCODING):
        """Context manager for direct use of PyFAT."""
        pf = PyFat(encoding=encoding, offset=offset)
        pf.open(filename)
        yield pf
        pf.close()



# -*- coding: utf-8 -*-
"""Registers the PyFatFSOpener used by PyFilesystem2."""

from fs.opener.parse import ParseResult

__all__ = ['PyFatFSOpener']

from fs.opener import Opener

from pyfatfs.PyFatFS import PyFatFS


class PyFatFSOpener(Opener):
    """Registers fat:// protocol for PyFilesystem2."""

    protocols = ['fat']

    @staticmethod
    def __param_parse(params: dict) -> dict:
        """Parse parameters and convert strings to bool."""
        _params = params.copy()
        for p in params:
            v = params[p]
            if isinstance(v, str):
                if v.lower() in ['true', '1', 't', 'y']:
                    v = True
                elif v.lower() in ['false', '0', 'f', 'n']:
                    v = False
            _params[p] = v
        return _params

    def open_fs(self, fs_url: str,  # pylint: disable=R0201
                parse_result: ParseResult, create: bool,
                cwd: str, writeable: bool = True):
        """Handle PyFilesystem2's protocol opening interface."""
        fs = PyFatFS(filename=parse_result.resource,
                     **self.__param_parse(parse_result.params))
        return fs



# -*- coding: utf-8 -*-

"""Implementation of `FatIO` for basic I/O."""
import errno
import io
import threading
from typing import Union, Optional

from fs.mode import Mode

from pyfatfs.PyFat import PyFat


class FatIO(io.RawIOBase):
    """Wrap basic I/O operations for PyFat."""

    def __init__(self, fs: PyFat,
                 path: str,
                 mode: Mode = Mode('r')) -> None:
        """Wrap basic I/O operations for PyFat. **Currently read-only**.

        :param fs: `PyFat`: Instance of opened filesystem
        :param path: `str`: Path to file. If `mode` is *r*,
                            the file must exist.
        :param mode: `Mode`: Mode to open file in.
        """
        super(FatIO, self).__init__()
        self.mode = mode
        self.fs = fs
        self.name = str(path)
        # TODO: File locking
        self._lock = threading.Lock()

        self.dir_entry = self.fs.root_dir.get_entry(path)
        if self.dir_entry.is_directory() or self.dir_entry.is_special():
            raise IsADirectoryError(errno.EISDIR, path)
        elif self.dir_entry.is_volume_id():
            raise FileNotFoundError(errno.ENOENT, path)

        #: Position in bytes from beginning of file
        self.__bpos = 0
        #: Current cluster chain number
        self.__cpos = self.dir_entry.get_cluster()
        #: Current cluster chain offset (in bytes)
        self.__coffpos = 0
        #: Handle to cluster chain iterator
        self.__fp = self.fs.get_cluster_chain(self.__cpos)

        if self.mode.appending:
            self.seek(0, 2)

    def __repr__(self) -> str:
        """Readable representation of class instance.

        ex: <FatFile fs=<PyFat object> path="/README.txt" mode="r">
        """
        return f'<{self.__class__.__name__} ' \
               f'fs={self.fs} ' \
               f'path="{self.name}" ' \
               f'mode="{self.mode}"'

    def seek(self, offset: int, whence: int = 0) -> int:
        """Seek to a given offset in the file.

        :param offset: ``int``: offset in bytes in the file
        :param whence: ``int``: offset position:
                       - ``0``: absolute
                       - ``1``: relative to current position
                       - ``2``: relative to file end
        :returns: New position in bytes in the file
        """
        if whence == 1:
            offset += self.__bpos
        elif whence == 2:
            offset += self.dir_entry.get_size()
        elif whence != 0:
            raise ValueError(f"Invalid whence {whence}, should be 0, 1 or 2")

        offset = min(offset, self.dir_entry.filesize)

        cluster_count = offset // self.fs.bytes_per_cluster
        self.__coffpos = offset % self.fs.bytes_per_cluster
        self.__bpos = offset
        self.__fp = self.fs.get_cluster_chain(self.dir_entry.get_cluster())
        self.__cpos = self.dir_entry.get_cluster()

        for _ in range(0, cluster_count):
            self.__cpos = next(self.__fp)

        return self.__bpos

    def seekable(self) -> bool:
        """FAT I/O driver is able to seek in files.

        :returns: `True`
        """
        return True

    def close(self) -> None:
        """Close open file handles assuming lock handle."""
        self.seek(0)
        self.fs.flush_fat()
        super().close()

    def readable(self) -> bool:
        """Determine whether or not the file is readable."""
        return self.mode.reading

    def read(self, size: int = -1) -> Union[bytes, None]:
        """Read given bytes from file."""
        if not self.mode.reading:
            raise IOError("File not open for reading")

        # Set size boundary
        if size + self.__bpos > self.dir_entry.filesize or size < 0:
            size = self.dir_entry.filesize - self.__bpos

        if size == 0:
            return b""

        chunks = []
        read_bytes = 0
        cluster_offset = self.__coffpos
        for c in self.__fp:
            chunk_size = self.fs.bytes_per_cluster - cluster_offset
            # Do not read past EOF
            if read_bytes + chunk_size > size:
                chunk_size = size - read_bytes

            chunk = self.fs.read_cluster_contents(c)
            chunk = chunk[cluster_offset:][:chunk_size]
            cluster_offset = 0
            chunks.append(chunk)
            read_bytes += chunk_size
            if read_bytes == size:
                break

        self.seek(read_bytes, 1)

        chunks = b"".join(chunks)
        if len(chunks) != size:
            raise RuntimeError("Read a different amount of data "
                               "than was requested.")
        return chunks

    def writable(self) -> bool:
        """Determine whether or not the file is writable."""
        if not self.dir_entry.is_read_only() and not self.fs.is_read_only:
            return self.mode.writing
        return False

    def write(self, __b: Union[bytes, bytearray]) -> Optional[int]:
        """Write given bytes to file."""
        if not self.writable():
            raise IOError('Cannot write to read-only file!')

        sz = len(__b)
        cluster = self.dir_entry.get_cluster()
        if sz == 0:
            # Nothing to do
            return sz
        elif cluster == 0:
            # Allocate new cluster chain if needed
            cluster = self.fs.allocate_bytes(sz)[0]
            self.dir_entry.set_cluster(cluster)
        else:
            # Get the current cluster and rewrite
            __coffpos = self.__coffpos
            if self.__bpos == self.dir_entry.filesize and self.__coffpos == 0:
                # New cluster required - concatenate to last cluster
                __coffpos = self.fs.bytes_per_cluster
                cluster = self.__cpos
            else:
                cluster = next(self.__fp)
            if __coffpos != 0:
                cluster_data = self.fs.read_cluster_contents(cluster)
                __b = cluster_data[0:__coffpos] + __b

        self.fs.write_data_to_cluster(__b, cluster)
        if self.__bpos + sz > self.dir_entry.filesize:
            self.dir_entry.filesize = self.__bpos + sz
        self.seek(sz, 1)
        self.fs.update_directory_entry(self.dir_entry.get_parent_dir())
        return sz

    def truncate(self, size: Optional[int] = 0) -> int:
        """Truncate file to given size.

        :param size: `int`: Size to truncate to, defaults to 0.
        :returns: `int`: Truncated size
        """
        cur_pos = self.tell()
        size = size if size is not None else cur_pos

        if size > self.dir_entry.get_size():
            self.seek(0, 2)
            self.write(b'\0' * (size - self.dir_entry.get_size()))
            self.seek(cur_pos)
        elif size < self.dir_entry.get_size():
            num_clusters = self.fs.calc_num_clusters(size)
            i = 0
            for c in self.fs.get_cluster_chain(self.dir_entry.get_cluster()):
                i += 1
                if i <= num_clusters:
                    continue
                self.fs.free_cluster_chain(c)
                self.fs.flush_fat()
                break

        # Update file size
        self.dir_entry.set_size(size)
        self.fs.update_directory_entry(self.dir_entry.get_parent_dir())
        return size



# -*- coding: utf-8 -*-

"""Directory entry operations with PyFAT."""
import posixpath
import struct

from pyfatfs.DosDateTime import DosDateTime
from pyfatfs.EightDotThree import EightDotThree
from pyfatfs._exceptions import PyFATException, NotAnLFNEntryException,\
    BrokenLFNEntryException
from pyfatfs import FAT_OEM_ENCODING, FAT_LFN_ENCODING

import errno


class FATDirectoryEntry:
    """Represents directory entries in FAT (files & directories)."""

    #: Bit set in DIR_Attr if entry is read-only
    ATTR_READ_ONLY = 0x01
    #: Bit set in DIR_Attr if entry is hidden
    ATTR_HIDDEN = 0x02
    #: Bit set in DIR_Attr if entry is a system file
    ATTR_SYSTEM = 0x04
    #: Bit set in DIR_Attr if entry is a volume id descriptor
    ATTR_VOLUME_ID = 0x8
    #: Bit set in DIR_Attr if entry is a directory
    ATTR_DIRECTORY = 0x10
    #: Bit set in DIR_Attr if entry is an archive
    ATTR_ARCHIVE = 0x20
    #: Bits set in DIR_Attr if entry is an LFN entry
    ATTR_LONG_NAME = ATTR_READ_ONLY | ATTR_HIDDEN | \
        ATTR_SYSTEM | ATTR_VOLUME_ID
    #: Bitmask to check if entry is an LFN entry
    ATTR_LONG_NAME_MASK = ATTR_READ_ONLY | ATTR_HIDDEN | ATTR_SYSTEM | \
        ATTR_VOLUME_ID | ATTR_DIRECTORY | ATTR_ARCHIVE

    #: Directory entry header layout in struct formatted string
    FAT_DIRECTORY_LAYOUT = "<11sBBBHHHHHHHL"
    #: Size of a directory entry header in bytes
    FAT_DIRECTORY_HEADER_SIZE = struct.calcsize(FAT_DIRECTORY_LAYOUT)
    #: Directory entry headers
    FAT_DIRECTORY_VARS = ["DIR_Name", "DIR_Attr", "DIR_NTRes",
                          "DIR_CrtTimeTenth", "DIR_CrtTime",
                          "DIR_CrtDate", "DIR_LstAccessDate",
                          "DIR_FstClusHI", "DIR_WrtTime",
                          "DIR_WrtDate", "DIR_FstClusLO",
                          "DIR_FileSize"]

    def __init__(self, DIR_Name, DIR_Attr, DIR_NTRes, DIR_CrtTimeTenth,
                 DIR_CrtTime, DIR_CrtDate, DIR_LstAccessDate,
                 DIR_FstClusHI, DIR_WrtTime, DIR_WrtDate, DIR_FstClusLO,
                 DIR_FileSize, encoding: str = FAT_OEM_ENCODING,
                 lfn_entry=None):
        """FAT directory entry constructor.

        :param DIR_Name: `EightDotThree` class instance
        :param DIR_Attr: Attributes of directory
        :param DIR_NTRes: Reserved attributes of directory entry
        :param DIR_CrtTimeTenth: Milliseconds at file creation
        :param DIR_CrtTime: Creation timestamp of entry
        :param DIR_CrtDate: Creation date of entry
        :param DIR_LstAccessDate: Last access date of entry
        :param DIR_FstClusHI: High cluster value of entry data
        :param DIR_WrtTime: Modification timestamp of entry
        :param DIR_WrtDate: Modification date of entry
        :param DIR_FstClusLO: Low cluster value of entry data
        :param DIR_FileSize: File size in bytes
        :param encoding: Encoding of filename
        :param lfn_entry: FATLongDirectoryEntry instance or None
        """
        self.name = DIR_Name
        self.attr = int(DIR_Attr)
        self.ntres = int(DIR_NTRes)
        self.crttimetenth = int(DIR_CrtTimeTenth)
        self.crttime = int(DIR_CrtTime)
        self.crtdate = int(DIR_CrtDate)
        self.lstaccessdate = int(DIR_LstAccessDate)
        self.fstclushi = int(DIR_FstClusHI)
        self.wrttime = int(DIR_WrtTime)
        self.wrtdate = int(DIR_WrtDate)
        self.fstcluslo = int(DIR_FstClusLO)
        self.filesize = int(DIR_FileSize)

        self._parent = None

        # Handle LFN entries
        self.lfn_entry = None
        self.set_lfn_entry(lfn_entry)

        self.__dirs = set()
        self.__encoding = encoding

    def get_ctime(self) -> DosDateTime:
        """Get dentry creation time."""
        return self.__combine_dosdatetime(self.crtdate, self.crttime)

    def get_mtime(self) -> DosDateTime:
        """Get dentry modification time."""
        return self.__combine_dosdatetime(self.wrtdate, self.wrttime)

    def get_atime(self) -> DosDateTime:
        """Get dentry access time."""
        return DosDateTime.deserialize_date(self.lstaccessdate)

    @staticmethod
    def __combine_dosdatetime(dt, tm) -> DosDateTime:
        dt = DosDateTime.deserialize_date(dt)
        dt.combine(dt, DosDateTime.deserialize_time(tm))
        return dt

    def get_checksum(self) -> int:
        """Get calculated checksum of this directory entry.

        :returns: Checksum as int
        """
        return self.name.checksum()

    def set_lfn_entry(self, lfn_entry):
        """Set LFN entry for current directory entry.

        :param: lfn_entry: Can be either of type `FATLongDirectoryEntry`
                or `None`.
        """
        if not isinstance(lfn_entry, FATLongDirectoryEntry):
            return

        # Verify LFN entries checksums
        chksum = self.get_checksum()
        for entry in lfn_entry.lfn_entries:
            entry_chksum = lfn_entry.lfn_entries[entry]["LDIR_Chksum"]
            if entry_chksum != chksum:
                raise BrokenLFNEntryException(f'Checksum verification for '
                                              f'LFN entry of directory '
                                              f'"{self.get_short_name()}" '
                                              f'failed')
        self.lfn_entry = lfn_entry

    def get_entry_size(self):
        """Get size of directory entry.

        :returns: Entry size in bytes as int
        """
        sz = self.FAT_DIRECTORY_HEADER_SIZE
        if isinstance(self.lfn_entry, FATLongDirectoryEntry):
            sz *= len(self.lfn_entry.lfn_entries)
        return sz

    def get_size(self):
        """Get filesize or directory entry size.

        :returns: Filesize or directory entry size in bytes as int
        """
        if self.is_directory():
            sz = self.FAT_DIRECTORY_HEADER_SIZE
            sz *= len(self.__dirs)+1
            return sz

        return self.filesize

    def set_size(self, size: int):
        """Set filesize.

        :param size: `int`: File size in bytes
        """
        self.filesize = size

    def get_cluster(self):
        """Get cluster address of directory entry.

        :returns: Cluster address of entry
        """
        return self.fstcluslo + (self.fstclushi << 16)

    def set_cluster(self, first_cluster):
        """Set low and high cluster address in directory headers."""
        self.fstcluslo = (first_cluster >> (16 * 0) & 0xFFFF)
        self.fstclushi = (first_cluster >> (16 * 1) & 0xFFFF)

    def byte_repr(self):
        """Represent directory entry as bytes.

        Note: Also represents accompanying LFN entries

        :returns: Entry & LFN entry as bytes-object
        """
        name = self.name.byte_repr()
        if name[0] == 0xE5:
            name[0] = 0x05

        entry = b''
        if isinstance(self.lfn_entry, FATLongDirectoryEntry):
            entry += self.lfn_entry.byte_repr()

        entry += struct.pack(self.FAT_DIRECTORY_LAYOUT, name, self.attr,
                             self.ntres, self.crttimetenth, self.crttime,
                             self.crtdate, self.lstaccessdate,
                             self.fstclushi, self.wrttime, self.wrtdate,
                             self.fstcluslo, self.filesize)

        return entry

    def _add_parent(self, cls):
        """Add parent directory link to current directory entry.

        raises: PyFATException
        """
        if self._parent is not None:
            raise PyFATException("Trying to add multiple parents to current "
                                 "directory!", errno=errno.ETOOMANYREFS)

        if not isinstance(cls, FATDirectoryEntry):
            raise PyFATException("Trying to add a non-FAT directory entry "
                                 "as parent directory!", errno=errno.EBADE)

        self._parent = cls

    def _get_parent_dir(self, sd):
        """Build path name for recursive directory entries."""
        name = self.__repr__()
        if self.__repr__() == "/":
            name = ""
        sd += [name]

        if self._parent is None:
            return sd

        return self._parent._get_parent_dir(sd)

    def get_full_path(self):
        """Iterate all parents up and join them by "/"."""
        parent_dirs = [self.__repr__()]

        if self._parent is None:
            return "/"

        return posixpath.join(list(reversed(
            self._parent._get_parent_dir(parent_dirs))))

    def get_parent_dir(self):
        """Get the parent directory entry."""
        if self._parent is None:
            raise PyFATException("Cannot query parent directory of "
                                 "root directory", errno=errno.ENOENT)

        return self._parent

    def is_special(self):
        """Determine if dir entry is a dot or dotdot entry.

        :returns: Boolean value whether or not entry is
                  a dot or dotdot entry
        """
        return self.get_short_name() in [".", ".."]

    def is_read_only(self):
        """Determine if dir entry has read-only attribute set.

        :returns: Boolean value indicating read-only attribute is set
        """
        return (self.ATTR_READ_ONLY & self.attr) > 0

    def is_hidden(self):
        """Determine if dir entry has the hidden attribute set.

        :returns: Boolean value indicating hidden attribute is set
        """
        return (self.ATTR_HIDDEN & self.attr) > 0

    def is_system(self):
        """Determine if dir entry has the system file attribute set.

        :returns: Boolean value indicating system attribute is set
        """
        return (self.ATTR_SYSTEM & self.attr) > 0

    def is_volume_id(self):
        """Determine if dir entry has the volume ID attribute set.

        :returns: Boolean value indicating volume ID attribute is set
        """
        return (self.ATTR_VOLUME_ID & self.attr) > 0

    def _verify_is_directory(self):
        """Verify that current entry is a directory.

        raises: PyFATException: If current entry is not a directory.
        """
        if not self.is_directory():
            raise PyFATException("Cannot get entries of this entry, as "
                                 "it is not a directory.",
                                 errno=errno.ENOTDIR)

    def is_directory(self):
        """Determine if dir entry has directory attribute set.

        :returns: Boolean value indicating directory attribute is set
        """
        return (self.ATTR_DIRECTORY & self.attr) > 0

    def is_archive(self):
        """Determine if dir entry has archive attribute set.

        :returns: Boolean value indicating archive attribute is set
        """
        return (self.ATTR_ARCHIVE & self.attr) > 0

    def is_empty(self):
        """Determine if directory does not contain any directories."""
        self._verify_is_directory()

        for d in self.__dirs:
            if d.is_special():
                continue
            return False

        return True

    def get_entries(self):
        """Get entries of directory.

        :raises: PyFatException: If entry is not a directory
        :returns: tuple: root (current path, full),
                 dirs (all dirs), files (all files)
        """
        dirs = []
        files = []
        specials = []

        self._verify_is_directory()

        for d in self.__dirs:
            if d.is_special() or d.is_volume_id():
                # Volume IDs and dot/dotdot entries
                specials += [d]
            elif d.is_directory():
                # Directories
                dirs += [d]
            else:
                # Everything else must be a file
                files += [d]

        return dirs, files, specials

    def _search_entry(self, name: str):
        """Find given dir entry by walking current dir.

        :param name: Name of entry to search for
        :raises: PyFATException: If entry cannot be found
        :returns: FATDirectoryEntry: Found entry
        """
        dirs, files, _ = self.get_entries()
        for entry in dirs+files:
            try:
                if entry.get_long_name() == name:
                    return entry
            except NotAnLFNEntryException:
                pass
            if entry.get_short_name() == name:
                return entry

        raise PyFATException(f'Cannot find entry {name}',
                             errno=errno.ENOENT)

    def get_entry(self, path: str):
        """Get sub-entry if current entry is a directory.

        :param path: Relative path of entry to get
        :raises: PyFATException: If entry cannot be found
        :returns: FATDirectoryEntry: Found entry
        """
        entry = self
        for segment in filter(None, path.split("/")):
            entry._verify_is_directory()
            entry = entry._search_entry(segment)
        return entry

    def walk(self):
        """Walk all directory entries recursively.

        :returns: tuple: root (current path, full),
                         dirs (all dirs), files (all files)
        """
        root = self.get_full_path()
        dirs, files, _ = self.get_entries()

        yield root, dirs, files
        for d in self.__dirs:
            if d.is_special():
                # Ignore dot and dotdot
                continue

            if not d.is_directory():
                continue

            yield from d.walk()

    def add_subdirectory(self, dir_entry):
        """Register a subdirectory in current directory entry.

        :param dir_entry: FATDirectoryEntry
        :raises: PyFATException: If current entry is not a directory or
                                 given directory entry already has a parent
                                 directory set
        """
        # Check if current dir entry is even a directory!
        self._verify_is_directory()

        dir_entry._add_parent(self)
        self.__dirs.add(dir_entry)

    def remove_dir_entry(self, name):
        """Remove given dir_entry from dir list.

        **NOTE:** This will also remove special entries such
        as ».«, »..« and volume labels!

        """
        d, f, s = self.get_entries()

        # Iterate all entries
        for dir_entry in d + f + s:
            sn = dir_entry.get_short_name()
            try:
                ln = dir_entry.get_long_name()
            except NotAnLFNEntryException:
                ln = None
            if name in [sn, ln]:
                self.__dirs.remove(dir_entry)
                return

        raise PyFATException(f"Cannot remove '{name}', no such "
                             f"file or directory!", errno=errno.ENOENT)

    def __repr__(self):
        """String-represent directory entry by (preferrably) LFN.

        :returns: str: Long file name if existing, 8DOT3 otherwise
        """
        try:
            return self.get_long_name()
        except NotAnLFNEntryException:
            return self.get_short_name()

    def get_short_name(self):
        """Get short name of directory entry.

        :returns: str: Name of directory entry
        """
        return self.name.get_unpadded_filename()

    def get_long_name(self):
        """Get long name of directory entry.

        :raises: NotAnLFNEntryException: If entry has no long file name
        :returns: str: Long file name of directory entry
        """
        if self.lfn_entry is None:
            raise NotAnLFNEntryException("No LFN entry found for this "
                                         "dir entry.")

        return str(self.lfn_entry)


class FATLongDirectoryEntry(object):
    """Represents long file name (LFN) entries."""

    #: LFN entry header layout in struct formatted string
    FAT_LONG_DIRECTORY_LAYOUT = "<B10sBBB12sH4s"
    #: LFN header fields when extracted with `FAT_LONG_DIRECTORY_LAYOUT`
    FAT_LONG_DIRECTORY_VARS = ["LDIR_Ord", "LDIR_Name1", "LDIR_Attr",
                               "LDIR_Type", "LDIR_Chksum", "LDIR_Name2",
                               "LDIR_FstClusLO", "LDIR_Name3"]
    #: Ordinance of last LFN entry in a chain
    LAST_LONG_ENTRY = 0x40
    #: Length for long file name in bytes per entry
    LFN_ENTRY_LENGTH = 26

    def __init__(self):
        """Initialize empty LFN directory entry object."""
        self.lfn_entries = {}

    def get_entries(self, reverse: bool = False):
        """Get LFS entries in correct order (based on `LDIR_Ord`).

        :param reverse: `bool`: Returns LFN entries in reversed order.
                        This is required for byte representation.
        """
        for _, e in sorted(self.lfn_entries.items(),
                           key=lambda x: x[1]["LDIR_Ord"],
                           reverse=reverse):
            yield e

    def byte_repr(self):
        """Represent LFN entries as bytes."""
        entries_bytes = b""
        for e in self.get_entries(reverse=True):
            entries_bytes += struct.pack(self.FAT_LONG_DIRECTORY_LAYOUT,
                                         e["LDIR_Ord"], e["LDIR_Name1"],
                                         e["LDIR_Attr"], e["LDIR_Type"],
                                         e["LDIR_Chksum"], e["LDIR_Name2"],
                                         e["LDIR_FstClusLO"], e["LDIR_Name3"])
        return entries_bytes

    def __str__(self):
        """Remove padding from LFN entry and decode it.

        :returns: `str` decoded string of filename
        """
        name = b''

        for e in self.get_entries():
            for h in ["LDIR_Name1", "LDIR_Name2", "LDIR_Name3"]:
                name += e[h]

        while name.endswith(b'\xFF\xFF'):
            name = name[:-2]

        name = name.decode(FAT_LFN_ENCODING)

        if name.endswith('\0'):
            name = name[:-1]

        return name

    @staticmethod
    def is_lfn_entry(LDIR_Ord, LDIR_Attr):
        """Verify that entry is an LFN entry.

        :param LDIR_Ord: First byte of the directory header, ordinance
        :param LDIR_Attr: Attributes segment of directory header
        :returns: `True` if entry is a valid LFN entry
        """
        lfn_attr = FATDirectoryEntry.ATTR_LONG_NAME
        lfn_attr_mask = FATDirectoryEntry.ATTR_LONG_NAME_MASK
        is_attr_set = (LDIR_Attr & lfn_attr_mask) == lfn_attr

        return is_attr_set and LDIR_Ord != 0xE5

    def add_lfn_entry(self, LDIR_Ord, LDIR_Name1, LDIR_Attr, LDIR_Type,
                      LDIR_Chksum, LDIR_Name2, LDIR_FstClusLO, LDIR_Name3):
        """Add LFN entry to this instances chain.

        :param LDIR_Ord: Ordinance of LFN entry
        :param LDIR_Name1: First name field of LFN entry
        :param LDIR_Attr: Attributes of LFN entry
        :param LDIR_Type: Type of LFN entry
        :param LDIR_Chksum: Checksum value of following 8dot3 entry
        :param LDIR_Name2: Second name field of LFN entry
        :param LDIR_FstClusLO: Cluster address of LFN entry. Always zero.
        :param LDIR_Name3: Third name field of LFN entry
        """
        # Check if attribute matches
        if not self.is_lfn_entry(LDIR_Ord, LDIR_Attr):
            raise NotAnLFNEntryException("Given LFN entry is not a long "
                                         "file name entry or attribute "
                                         "not set correctly!")

        # Check if FstClusLO is 0, as required by the spec
        if LDIR_FstClusLO != 0:
            raise PyFATException("Given LFN entry has an invalid first "
                                 "cluster ID, don't know what to do.",
                                 errno=errno.EFAULT)

        # Check if item with same index has already been added
        if LDIR_Ord in self.lfn_entries.keys():
            raise PyFATException("Given LFN entry part with index \'{}\'"
                                 "has already been added to LFN "
                                 "entry list.".format(LDIR_Ord))

        mapped_entries = dict(zip(self.FAT_LONG_DIRECTORY_VARS,
                                  (LDIR_Ord, LDIR_Name1, LDIR_Attr, LDIR_Type,
                                   LDIR_Chksum, LDIR_Name2, LDIR_FstClusLO,
                                   LDIR_Name3)))

        self.lfn_entries[LDIR_Ord] = mapped_entries

    def is_lfn_entry_complete(self):
        """Verify that LFN object forms a complete chain.

        :returns: `True` if `LAST_LONG_ENTRY` is found
        """
        for k in self.lfn_entries.keys():
            if (int(k) & self.LAST_LONG_ENTRY) == self.LAST_LONG_ENTRY:
                return True

        return False


def make_lfn_entry(dir_name: str,
                   short_name):
    """Generate a `FATLongDirectoryEntry` instance from directory name.

    :param dir_name: Long name of directory
    :param short_name: `EightDotThree` class instance
    :raises: `PyFATException` if entry name does not require an LFN
             entry or the name exceeds the FAT limitation of 255 characters
    """
    lfn_entry = FATLongDirectoryEntry()
    #: Length in bytes of an LFN entry
    lfn_entry_length = 26
    dir_name_str = dir_name
    dir_name = dir_name.encode(FAT_LFN_ENCODING)
    dir_name_modulus = len(dir_name) % lfn_entry_length

    if EightDotThree.is_8dot3_conform(dir_name_str):
        raise PyFATException("Directory entry is already 8.3 conform, "
                             "no need to create an LFN entry.",
                             errno=errno.EINVAL)

    if len(dir_name) > 255:
        raise PyFATException("Long file name exceeds 255 "
                             "characters, not supported.",
                             errno=errno.ENAMETOOLONG)

    checksum = short_name.checksum()

    if dir_name_modulus != 0:
        # Null-terminate string if required
        dir_name += '\0'.encode(FAT_LFN_ENCODING)

    # Fill the rest with 0xFF if it doesn't fit evenly
    new_sz = lfn_entry_length - len(dir_name)
    new_sz %= lfn_entry_length
    new_sz += len(dir_name)
    dir_name += b'\xFF' * (new_sz - len(dir_name))

    # Generate linked LFN entries
    lfn_entries = len(dir_name) // lfn_entry_length
    for i in range(lfn_entries):
        if i == lfn_entries-1:
            lfn_entry_ord = 0x40 | i+1
        else:
            lfn_entry_ord = i+1

        n = i*lfn_entry_length
        dirname1 = dir_name[n:n+10]
        n += 10
        dirname2 = dir_name[n:n+12]
        n += 12
        dirname3 = dir_name[n:n+4]

        lfn_entry.add_lfn_entry(LDIR_Ord=lfn_entry_ord,
                                LDIR_Name1=dirname1,
                                LDIR_Attr=FATDirectoryEntry.ATTR_LONG_NAME,
                                LDIR_Type=0x00,
                                LDIR_Chksum=checksum,
                                LDIR_Name2=dirname2,
                                LDIR_FstClusLO=0,
                                LDIR_Name3=dirname3)
    return lfn_entry



# -*- coding: utf-8 -*-

"""Tests from PyFilesystem2."""
import gzip
import os
from unittest import TestCase

from fs.test import FSTestCases

from pyfatfs.PyFatFS import PyFatFS


class TestPyFatFS16(FSTestCases, TestCase):
    """Integration tests with PyFilesystem2 for FAT16."""

    def make_fs(self):  # pylint: disable=R0201
        """Create filesystem for pyfilesystem2 integration tests."""
        img_file = os.path.join(os.path.dirname(__file__),
                                "data", "pyfat16.img")
        with gzip.open(img_file + '.gz', "r") as imggz:
            with open(img_file, 'wb') as img:
                img.write(imggz.read())

        return PyFatFS(img_file, encoding='UTF-8')



# -*- coding: utf-8 -*-

"""Custom Exceptions for PyFAT."""


class PyFATException(Exception):
    """Generic PyFAT Exceptions."""

    def __init__(self, msg: str, errno=None):
        """Construct base class for PyFAT exceptions.

        :param msg: Exception message describing what happened
        :param errno: Error number, mostly based on POSIX errno where feasible
        """
        Exception.__init__(self, msg)
        self.errno = errno


class NotAnLFNEntryException(PyFATException):
    """Indicates that given dir entry cannot be interpreted as LFN entry."""


class BrokenLFNEntryException(PyFATException):
    """Indicates that given LFN entry is invalid."""


class NotAFatEntryException(NotADirectoryError):
    """Custom handling for FAT `NotADirectoryError`'s."""

    #: Indicates a free entry, but not an end of chain.
    FREE_ENTRY = 0xE5
    #: Indicates an end of directory cluster, do not search further.
    LAST_ENTRY = 0x00

    def __init__(self, msg: str, free_type=FREE_ENTRY):
        """Construct base class for PyFAT exceptions.

        :param msg: Exception message describing what happened
        :param free_type: Either `FREE_ENTRY` or `LAST_ENTRY`.
        """
        NotADirectoryError.__init__(self, msg)
        self.free_type = free_type



# -*- coding: utf-8 -*-

"""
Python FAT filesystem module with :doc:`PyFilesystem2 <pyfilesystem2:index>` \
compatibility.

pyfatfs allows interaction with FAT12/16/32 filesystems, either via
:doc:`PyFilesystem2 <pyfilesystem2:index>` for file-level abstraction
or direct interaction with the filesystem for low-level access.
"""

__name__ = 'pyfatfs'
__author__ = 'Nathan-J. Hirschauer'
__author_email__ = 'nathanhi@deepserve.info'
__license__ = 'MIT License'


#: Specifies default ("OEM") encoding
FAT_OEM_ENCODING = 'ibm437'
#: Specifies the long file name encoding, which is always UTF-16 (LE)
FAT_LFN_ENCODING = 'utf-16-le'



# -*- coding: utf-8 -*-

"""Tests the 8DOT3 short name module."""
import string
import random
from unittest import mock

import pytest

from pyfatfs.EightDotThree import EightDotThree


def test_bytename_set_invalid_length():
    """Test that a byte name can only be set with the correct length."""
    sfn = EightDotThree()
    for n in [b'', b'1234567890', b'123456789011']:
        with pytest.raises(ValueError):
            sfn.set_byte_name(n)


def test_bytename_set_not_a_directory_last_entry():
    """Test that 0x00 is detect as the last entry of the directory cluster."""
    sfn = EightDotThree()
    with pytest.raises(NotADirectoryError) as ex:
        sfn.set_byte_name(b'\0          ')
        assert ex.free_type == ex.LAST_ENTRY


def test_bytename_set_not_a_directory_free_entry():
    """Test that 0x00 is detect as the last entry of the directory cluster."""
    sfn = EightDotThree()
    with pytest.raises(NotADirectoryError) as ex:
        sfn.set_byte_name(b'\xE5          ')
        assert ex.free_type == ex.FREE_ENTRY


def test_checksum_calculation_precalculated():
    """Test that the checksum calculation works, with a precalculated value."""
    sfn = EightDotThree()
    sfn.set_str_name("FILENAME.TXT")
    assert sfn.checksum() == 58


def calculate_checksum_referenceimpl(fname: bytes) -> int:
    """Calculate checksum of a string based on reference implementation.

    This is taken from the Microsoft Extensible Firmware Initiative
    FAT32 File System Spec v1.03.

    .. code-block:
       short FcbNameLen;
       unsigned char Sum;

       Sum = 0;
       for (FcbNameLen=11; FcbNameLen!=0; FcbNameLen--) {
           Sum = ((Sum & 1) ? 0x80 : 0) + (Sum >> 1) + *pFcbName++;
       }

    """
    s = 0
    for i in range(0, 11):
        s = (0x80 if (s & 1) == 1 else 0x0) + (s >> 1) + fname[i]
        s &= 0xFF

    return s


def test_checksum_calculation_referenceimpl():
    """Test that the checksum calculation works, following reference impl."""
    sfn = EightDotThree()
    sfn.set_str_name("FILENAME.TXT")
    assert sfn.checksum() == calculate_checksum_referenceimpl(sfn.name)


def test_checksum_calculation_referenceimpl_random():
    """Test that the checksum calculation works, following reference impl."""
    base = ''.join(random.choices(string.ascii_letters + string.digits,
                                  k=8)).upper()
    ext = ''.join(random.choices(string.ascii_letters + string.digits,
                                 k=3)).upper()
    sfn = EightDotThree()
    sfn.set_str_name(f"{base}.{ext}")

    assert sfn.checksum() == calculate_checksum_referenceimpl(sfn.name)


def test_make_8dot3_name():
    """Test that make_8dot3_filename generates valid 8dot3 filenames."""
    fde = mock.MagicMock()
    fde.get_entries.return_value = ([], [], [])
    sfn = EightDotThree()
    n = sfn.make_8dot3_name("This is a long filename.txt", fde)
    sfn.set_str_name(n)
    assert "THISIS.TXT" == n
    assert sfn.is_8dot3_conform(sfn.get_unpadded_filename())


def test_make_8dot3_name_cut_ext():
    """Test that make_8dot3_filename generates valid 8dot3 filenames."""
    fde = mock.MagicMock()
    fde.get_entries.return_value = ([], [], [])
    sfn = EightDotThree()
    lfn = sfn.make_8dot3_name("This is a long filename.TeXT", fde)
    assert "THISIS.TEX" == lfn
    assert sfn.is_8dot3_conform(lfn)


def test_make_8dot3_name_noext():
    """Test that make_8dot3_filename generates valid 8dot3 filenames."""
    fde = mock.MagicMock()
    fde.get_entries.return_value = ([], [], [])
    sfn = EightDotThree()
    lfn = sfn.make_8dot3_name("This is a long filename", fde)
    assert "THISIS" == lfn
    assert sfn.is_8dot3_conform(lfn)


def test_make_8dot3_name_emptyext():
    """Test that make_8dot3_filename generates valid 8dot3 filenames."""
    fde = mock.MagicMock()
    fde.get_entries.return_value = ([], [], [])
    sfn = EightDotThree()
    lfn = sfn.make_8dot3_name("This is a long filename.", fde)
    assert "THISIS" == lfn
    assert sfn.is_8dot3_conform(lfn)


def test_make_8dot3_name_unicode():
    """Test that make_8dot3_filename generates valid 8dot3 filenames."""
    fde = mock.MagicMock()
    fde.get_entries.return_value = ([], [], [])
    sfn = EightDotThree()
    lfn = sfn.make_8dot3_name("🤷.🤷", fde)
    assert "_._" == lfn
    assert sfn.is_8dot3_conform(lfn)


def test_make_8dot3_name_collision():
    """Test that make_8dot3_filename generates valid 8dot3 filenames."""
    fde = mock.MagicMock()
    fde_sub = mock.MagicMock()
    fde_sub.get_short_name.side_effect = ["THISIS.TXT", "THISIS~1.TXT",
                                          "THISIS~2.TXT"]
    fde.get_entries.return_value = ([fde_sub, fde_sub], [fde_sub], [])
    sfn = EightDotThree()
    lfn = sfn.make_8dot3_name("This is a long filename.txt", fde)
    assert "THISIS~3.TXT" == lfn
    assert sfn.is_8dot3_conform(lfn)


def test_is_8dot3_conform_true():
    """Test that 8.3 file names are correctly detected."""
    assert EightDotThree.is_8dot3_conform("TESTFILE.TXT")


def test_is_8dot3_conform_false():
    """Test that non-8.3 file names are correctly detected."""
    assert not EightDotThree.is_8dot3_conform("This is a Long file.txt")



# -*- coding: utf-8 -*-

"""PyFilesystem2 implementation of PyFAT."""
import datetime
import posixpath
import errno

from fs.base import FS
from fs.mode import Mode
from fs.path import split, normpath
from fs.permissions import Permissions
from fs.info import Info
from fs.errors import DirectoryExpected, DirectoryExists, \
    ResourceNotFound, FileExpected, DirectoryNotEmpty, RemoveRootError, \
    FileExists
from fs import ResourceType
from fs.subfs import SubFS

from pyfatfs import FAT_OEM_ENCODING
from pyfatfs.DosDateTime import DosDateTime
from pyfatfs.PyFat import PyFat
from pyfatfs.FATDirectoryEntry import FATDirectoryEntry, make_lfn_entry
from pyfatfs._exceptions import PyFATException
from pyfatfs.FatIO import FatIO
from pyfatfs.EightDotThree import EightDotThree


class PyFatFS(FS):
    """PyFilesystem2 extension."""

    def __init__(self, filename: str, encoding: str = FAT_OEM_ENCODING,
                 offset: int = 0, preserve_case: bool = True,
                 read_only: bool = False, utc: bool = False):
        """PyFilesystem2 FAT constructor, initializes self.fs.

        :param filename: `str`: Name of file/device to open as FAT partition.
        :param encoding: `str`: Valid Python standard encoding.
        :param offset: `int`: Offset from file start to filesystem
                       start in bytes.
        :param preserve_case: `bool`: By default 8DOT3 filenames do not
                              support casing. If preserve_case is set to
                              `True`, it will create an LFN entry if the
                              casing does not conform to 8DOT3.
        :param read_only: `bool`: If set to true, the filesystem is mounted
                          in read-only mode, not allowing any modifications.
        :param utc: `bool`: Store timestamps in UTC rather than the local time.
                    This applies to dentry creation, modification and last
                    access time.
        """
        super(PyFatFS, self).__init__()
        self.preserve_case = preserve_case
        self.fs = PyFat(encoding=encoding, offset=int(offset))
        self.fs.open(filename, read_only=read_only)

        if utc:
            self.tz = datetime.timezone.utc
        else:
            self.tz = datetime.datetime.now(datetime.timezone.utc)
            self.tz = self.tz.astimezone().tzinfo

    def close(self):
        """Clean up open handles."""
        try:
            self.fs.close()
        except PyFATException:
            # Ignore if filesystem is already closed
            pass

        super(PyFatFS, self).close()

    def exists(self, path: str):
        """Verify if given path exists on filesystem.

        :param path: Path to file or directory on filesystem
        :returns: Boolean value indicating entries existence
        """
        try:
            self.fs.root_dir.get_entry(path)
        except PyFATException as e:
            if e.errno == errno.ENOENT:
                return False
            raise e

        return True

    def getinfo(self, path: str, namespaces=None):
        """Generate PyFilesystem2's `Info` struct.

        :param path: Path to file or directory on filesystem
        :param namespaces: Info namespaces to query, `NotImplemented`
        :returns: `Info`
        """
        try:
            entry = self.fs.root_dir.get_entry(path)
        except PyFATException as e:
            if e.errno in [errno.ENOTDIR, errno.ENOENT]:
                raise ResourceNotFound(path)
            raise e

        info = {"basic": {"name": repr(entry),
                          "is_dir": entry.is_directory()},
                "details": {"accessed": entry.get_atime().timestamp(),
                            "created": entry.get_ctime().timestamp(),
                            "metadata_changed": None,
                            "modified": entry.get_mtime().timestamp(),
                            "size": entry.filesize,
                            "type": self.gettype(path)}}
        return Info(info)

    def getmeta(self, namespace=u'standard'):
        """Get generic filesystem metadata.

        :param namespace: Namespace to query, only `standard` supported
        :returns: `dict` with file system meta data
        """
        if namespace != u'standard':
            return {}

        return {"case_insensitive": not self.preserve_case,
                "invalid_path_chars": "\0",
                "max_path_length": 255,
                "max_sys_path": None,
                "network": False,
                "read_only": self.fs.is_read_only,
                "unicode_paths": self.fs.encoding.lower().startswith('utf'),
                "supports_rename": True}

    def getsize(self, path: str):
        """Get size of file in bytes.

        :param path: Path to file or directory on filesystem
        :returns: Size in bytes as `int`
        """
        try:
            entry = self.fs.root_dir.get_entry(path)
        except PyFATException as e:
            if e.errno == errno.ENOENT:
                raise ResourceNotFound(path)
            raise e
        return entry.filesize

    def gettype(self, path: str):
        """Get type of file as `ResourceType`.

        :param path: Path to file or directory on filesystem
        :returns: `ResourceType.directory` or `ResourceType.file`
        """
        entry = self.fs.root_dir.get_entry(path)
        if entry.is_directory():
            return ResourceType.directory

        return ResourceType.file

    def listdir(self, path: str):
        """List contents of given directory entry.

        :param path: Path to directory on filesystem
        """
        dir_entry = self._get_dir_entry(path)
        try:
            dirs, files, _ = dir_entry.get_entries()
        except PyFATException as e:
            if e.errno == errno.ENOTDIR:
                raise DirectoryExpected(path)
            raise e
        return [str(e) for e in dirs+files]

    def create(self, path: str, wipe: bool = False) -> bool:
        """Create a new file.

        :param path: Path of new file on filesystem
        :param wipe: Overwrite existing file contents
        """
        basename = "/".join(path.split("/")[:-1])
        dirname = path.split("/")[-1]

        # Plausability checks
        try:
            self.opendir(basename)
        except DirectoryExpected:
            raise ResourceNotFound(path)
        base = self._get_dir_entry(basename)

        try:
            dentry = self._get_dir_entry(path)
        except ResourceNotFound:
            pass
        else:
            if not wipe:
                return False
            else:
                # Clean up existing file contents
                # TODO: touch {a,m}time
                dentry.set_size(0)
                old_cluster = dentry.get_cluster()
                dentry.set_cluster(0)
                self.fs.free_cluster_chain(old_cluster)
                return True

        # Determine 8DOT3 file name + LFN
        short_name = EightDotThree()
        n = short_name.make_8dot3_name(dirname, base)
        short_name.set_str_name(n)

        dt = DosDateTime.now(tz=self.tz)

        newdir = FATDirectoryEntry(DIR_Name=short_name,
                                   DIR_Attr=0,
                                   DIR_NTRes=0,
                                   DIR_CrtTimeTenth=0,
                                   DIR_CrtTime=dt.serialize_time(),
                                   DIR_CrtDate=dt.serialize_date(),
                                   DIR_LstAccessDate=dt.serialize_date(),
                                   DIR_FstClusHI=0x00,
                                   DIR_WrtTime=dt.serialize_time(),
                                   DIR_WrtDate=dt.serialize_date(),
                                   DIR_FstClusLO=0x00,
                                   DIR_FileSize=0,
                                   encoding=self.fs.encoding)

        # Create LFN entry if required
        _sfn = short_name.get_unpadded_filename()
        if _sfn != dirname.upper() or (_sfn != dirname and self.preserve_case):
            lfn_entry = make_lfn_entry(dirname, short_name)
            newdir.set_lfn_entry(lfn_entry)

        # Write reference to parent directory
        base.add_subdirectory(newdir)
        self.fs.update_directory_entry(base)

        # Flush FAT(s) to disk
        self.fs.flush_fat()
        return True

    def makedir(self, path: str, permissions: Permissions = None,
                recreate: bool = False):
        """Create directory on filesystem.

        :param path: Path of new directory on filesystem
        :param permissions: Currently not implemented
        :param recreate: Ignore if directory already exists
        """
        path = normpath(path)
        base = split(path)[0]
        dirname = split(path)[1]

        # Plausability checks
        try:
            self.opendir(base)
        except DirectoryExpected:
            raise ResourceNotFound(path)
        base = self._get_dir_entry(base)

        try:
            self._get_dir_entry(path)
        except ResourceNotFound:
            pass
        else:
            if not recreate:
                raise DirectoryExists(path)
            else:
                # TODO: Update mtime
                return SubFS(self, path)

        parent_is_root = base == self.fs.root_dir

        # Determine 8DOT3 file name + LFN
        short_name = EightDotThree()
        n = short_name.make_8dot3_name(dirname, base)
        short_name.set_str_name(n)

        dt = DosDateTime.now(tz=self.tz)

        newdir = FATDirectoryEntry(DIR_Name=short_name,
                                   DIR_Attr=FATDirectoryEntry.ATTR_DIRECTORY,
                                   DIR_NTRes=0,
                                   DIR_CrtTimeTenth=0,
                                   DIR_CrtTime=dt.serialize_time(),
                                   DIR_CrtDate=dt.serialize_date(),
                                   DIR_LstAccessDate=dt.serialize_date(),
                                   DIR_FstClusHI=0x00,
                                   DIR_WrtTime=dt.serialize_time(),
                                   DIR_WrtDate=dt.serialize_date(),
                                   DIR_FstClusLO=0x00,
                                   DIR_FileSize=0,
                                   encoding=self.fs.encoding)

        # Create LFN entry if required
        _sfn = short_name.get_unpadded_filename()
        if _sfn != dirname.upper() or (_sfn != dirname and self.preserve_case):
            lfn_entry = make_lfn_entry(dirname, short_name)
            newdir.set_lfn_entry(lfn_entry)

        # Create . and .. directory entries
        first_cluster = self.fs.allocate_bytes(
            FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE * 2,
            erase=True)[0]
        newdir.set_cluster(first_cluster)
        dot_sn = EightDotThree()
        dot_sn.set_byte_name(b".          ")
        dot = FATDirectoryEntry(DIR_Name=dot_sn,
                                DIR_Attr=FATDirectoryEntry.ATTR_DIRECTORY,
                                DIR_NTRes=newdir.ntres,
                                DIR_CrtTimeTenth=newdir.crttimetenth,
                                DIR_CrtTime=newdir.crttime,
                                DIR_CrtDate=newdir.crtdate,
                                DIR_LstAccessDate=newdir.lstaccessdate,
                                DIR_FstClusHI=newdir.fstclushi,
                                DIR_WrtTime=newdir.wrttime,
                                DIR_WrtDate=newdir.wrtdate,
                                DIR_FstClusLO=newdir.fstcluslo,
                                DIR_FileSize=newdir.filesize,
                                encoding=self.fs.encoding)
        dotdot_sn = EightDotThree()
        dotdot_sn.set_byte_name(b"..         ")
        base_fstclushi = base.fstclushi if not parent_is_root else 0x0
        base_fstcluslo = base.fstcluslo if not parent_is_root else 0x0
        dotdot = FATDirectoryEntry(DIR_Name=dotdot_sn,
                                   DIR_Attr=FATDirectoryEntry.ATTR_DIRECTORY,
                                   DIR_NTRes=base.ntres,
                                   DIR_CrtTimeTenth=base.crttimetenth,
                                   DIR_CrtTime=base.crttime,
                                   DIR_CrtDate=base.crtdate,
                                   DIR_LstAccessDate=base.lstaccessdate,
                                   DIR_FstClusHI=base_fstclushi,
                                   DIR_WrtTime=base.wrttime,
                                   DIR_WrtDate=base.wrtdate,
                                   DIR_FstClusLO=base_fstcluslo,
                                   DIR_FileSize=base.filesize,
                                   encoding=self.fs.encoding)
        newdir.add_subdirectory(dot)
        newdir.add_subdirectory(dotdot)

        # Write new directory contents
        self.fs.update_directory_entry(newdir)

        # Write parent directory
        base.add_subdirectory(newdir)
        self.fs.update_directory_entry(base)

        # Flush FAT(s) to disk
        self.fs.flush_fat()

        return SubFS(self, path)

    def removedir(self, path: str):
        """Remove empty directories from the filesystem.

        :param path: `str`: Directory to remove
        """
        dir_entry = self._get_dir_entry(path)
        try:
            base = dir_entry.get_parent_dir()
        except PyFATException as e:
            if e.errno == errno.ENOENT:
                # Don't remove root directory
                raise RemoveRootError(path)
            raise e

        # Verify if the directory is empty
        try:
            if not dir_entry.is_empty():
                raise DirectoryNotEmpty(path)
        except PyFATException as e:
            if e.errno == errno.ENOTDIR:
                raise DirectoryExpected(path)

        self._remove(base, dir_entry)

    def removetree(self, dir_path: str):
        """Recursively remove the contents of a directory.

        :param dir_path: ``str``: Path to a directory on the filesystem.
        """
        dir_entry = self._get_dir_entry(dir_path)

        if not dir_entry.is_directory():
            raise DirectoryExpected(dir_path)

        dirs, files, _ = dir_entry.get_entries()

        for f in files:
            self._remove(dir_entry, f)

        for d in dirs:
            self.removetree(posixpath.join(dir_path, str(d)))

        return self.removedir(dir_path)

    def remove(self, path: str):
        """Remove a file from the filesystem.

        :param path: `str`: Path of file to remove
        """
        dir_entry = self._get_dir_entry(path)

        # Check for file
        if dir_entry.is_directory() or dir_entry.is_special():
            raise FileExpected(path)

        base = dir_entry.get_parent_dir()
        self._remove(base, dir_entry)

    def _remove(self, parent_dir: FATDirectoryEntry,
                dir_entry: FATDirectoryEntry):
        """Remove directory entry regardless of type (dir or file).

        **NOTE:** This will not recursively remove directories, thus
        leave allocated clusters behind unless the directory has been
        purged before. Also, there is no check for special files such as
        volume labels, ».« and »..« entries. So it might leave the filesystem
        in a broken state if used incorrectly.

        :param parent_dir: ``FATDirectoryEntry``: Parent directory
        :param dir_entry: ``FATDirectoryEntry``: Directory entry to remove
        :raises PyFATException: ``ENOENT`` if given dir entry does not exist
                                in ``parent_dir``
        """
        # Remove entry from parent directory
        parent_dir.remove_dir_entry(str(dir_entry))
        self.fs.update_directory_entry(parent_dir)

        # Free cluster in FAT
        if dir_entry.get_size() > 0 and dir_entry.get_cluster() != 0:
            # Empty files have a cluster ID of 0
            self.fs.free_cluster_chain(dir_entry.get_cluster())
        del dir_entry

    def openbin(self, path: str, mode: str = "r",
                buffering: int = -1, **options):
        """Open file from filesystem.

        :param path: Path to file on filesystem
        :param mode: Mode to open file in
        :param buffering: TBD
        returns: `BinaryIO` stream
        """
        path = self.validatepath(path)
        mode = Mode(mode)
        if mode.create:
            if mode.exclusive:
                try:
                    self.getinfo(path)
                except ResourceNotFound:
                    pass
                else:
                    raise FileExists(path)
            self.create(path)
        if "t" in mode:
            raise ValueError('Text-mode not allowed in openbin')

        try:
            info = self.getinfo(path)
        except ResourceNotFound:
            raise ResourceNotFound(path)
        else:
            if info.is_dir:
                raise FileExpected(path)

        return FatIO(self.fs, path, mode)

    def _get_dir_entry(self, path: str) -> FATDirectoryEntry:
        """Get a filesystem object for a path.

        :param path: `str`: Path on the filesystem
        :returns: `FATDirectoryEntry`
        """
        _path = normpath(self.validatepath(path))
        try:
            dir_entry = self.fs.root_dir.get_entry(_path)
        except PyFATException as e:
            if e.errno == errno.ENOENT:
                raise ResourceNotFound(path)
            raise e

        return dir_entry

    def opendir(self, path: str, factory=None) -> SubFS:
        """Get a filesystem object for a sub-directory.

        :param path: str: Path to a directory on the filesystem.
        """
        factory = factory or self.subfs_class or SubFS

        dir_entry = self._get_dir_entry(path)

        if not dir_entry.is_directory():
            raise DirectoryExpected(path)

        return factory(self, path)

    def setinfo(self, path: str, info):
        """Not yet properly implemented."""
        details = info.get('details', {})
        dentry = self._get_dir_entry(path)

        ctime = details.get("created")
        mtime = details.get("modified")
        atime = details.get("accessed")
        if ctime:
            ctime = DosDateTime.fromtimestamp(ctime, tz=self.tz)
            dentry.crttime = ctime.serialize_time()
            dentry.crtdate = ctime.serialize_date()
        if mtime:
            mtime = DosDateTime.fromtimestamp(mtime, tz=self.tz)
            dentry.wrttime = mtime.serialize_time()
            dentry.wrtdate = mtime.serialize_date()
        if atime:
            atime = DosDateTime.fromtimestamp(atime, tz=self.tz)
            dentry.lstaccessdate = atime.serialize_date()

        self.fs.update_directory_entry(dentry.get_parent_dir())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""pyFAT package definition."""

import io
import re

from setuptools import setup, find_packages

try:
    # pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:
    # pip <= 9.0.3
    from pip.req import parse_requirements


def load_requirements(fname):
    reqs = parse_requirements(fname, session="test")
    return [getattr(r, 'requirement',
            str(getattr(r, 'req', None))) for r in reqs]


def _get_attribute(name):
    """Get version information from __init__.py."""
    with io.open('pyfatfs/__init__.py') as f:
        return re.search(r"{}\s*=\s*'([^']+)'\s*".format(name),
                         f.read()).group(1)


def _get_readme():
    """Get contents of README.rst."""
    with io.open('README.rst') as readme:
        return readme.read()


setup(name=_get_attribute('__name__'),
      use_scm_version=True,
      description='FAT12/16/32 implementation with VFAT support',
      long_description=_get_readme(),
      long_description_content_type='text/x-rst',
      author=_get_attribute('__author__'),
      author_email=_get_attribute('__author_email__'),
      license=_get_attribute('__license__'),
      url='https://github.com/nathanhi/pyfatfs',
      project_urls={'Documentation': f'https://pyfatfs.readthedocs.io',
                    'Changelog': 'https://github.com/nathanhi/pyfatfs/blob/master/CHANGELOG.rst',
                    'Issues': 'https://github.com/nathanhi/pyfatfs/issues',
                    'Source': 'https://github.com/nathanhi/pyfatfs',
      },
      packages=find_packages(),
      keywords=['filesystem', 'PyFilesystem2', 'FAT12',
                'FAT16', 'FAT32', 'VFAT', 'LFN'],
      python_requires='~=3.6',
      test_suite='tests',
      install_requires=load_requirements("requirements/install.txt"),
      setup_requires=['pytest-runner', 'setuptools_scm~=5.0.0'],
      tests_require=load_requirements("requirements/test.txt"),
      entry_points={
          'fs.opener': ['fat = pyfatfs.PyFatFSOpener:PyFatFSOpener'],
      },
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: System :: Filesystems'],
      )



# -*- coding: utf-8 -*-

"""Enhancement of datetime for DOS date/time format compatibility."""
from datetime import datetime, time


def _convert_to_dos_date(func):
    """Convert datetime to DosDateTime."""
    def _wrapper(*args, **kwargs) -> "DosDateTime":
        date: datetime = func(*args, **kwargs)
        return DosDateTime(*date.timetuple()[:6])
    return _wrapper


class DosDateTime(datetime):
    """DOS-specific date/time format serialization."""

    now = _convert_to_dos_date(datetime.now)
    fromtimestamp = _convert_to_dos_date(datetime.fromtimestamp)

    def serialize_date(self) -> int:
        """Convert current datetime to FAT date."""
        date = self.year - 1980 << 9 | self.month << 5 | self.day
        return date

    def serialize_time(self) -> int:
        """Convert current datetime to FAT time."""
        time = self.hour << 11 | self.minute << 5
        time |= ((self.second - (self.second % 2)) // 2)
        return time

    @staticmethod
    def deserialize_date(dt: int) -> "DosDateTime":
        """Convert a DOS date format to a Python object."""
        day = dt & ((1 << 5) - 1)
        month = (dt >> 5) & ((1 << 4) - 1)
        year = ((dt >> 9) & (1 << 8) - 1) + 1980

        try:
            return DosDateTime(year, month, day)
        except ValueError:
            return DosDateTime(1980, 1, 1)

    @staticmethod
    def deserialize_time(tm: int) -> time:
        """Convert a DOS time format to a Python object."""
        second = (tm & (1 << 5) - 1) * 2
        minute = (tm >> 5) & ((1 << 5) - 1)
        hour = (tm >> 11) & ((1 << 5) - 1)
        return time(hour, minute, second)



# -*- coding: utf-8 -*-

"""8DOT3 file name helper class & functions."""

import errno
import os
import string

from pyfatfs import FAT_OEM_ENCODING
from pyfatfs._exceptions import PyFATException, NotAFatEntryException


def _init_check(func):
    def _wrapper(*args, **kwargs):
        initialised = args[0].initialised

        if initialised is True:
            return func(*args, **kwargs)
        else:
            raise PyFATException("Class has not yet been fully initialised, "
                                 "please instantiate first.")

    return _wrapper


class EightDotThree:
    """8DOT3 filename representation."""

    #: Length of the byte representation in a directory entry header
    SFN_LENGTH = 11

    #: Valid characters for 8.3 file names
    VALID_CHARACTERS = f"{string.ascii_letters.upper()}{string.digits}" \
                       "!#$%&'()-@^_`{}~"

    def __init__(self, encoding: str = FAT_OEM_ENCODING):
        """Offer 8DOT3 filename operation.

        :param encoding: Codepage for the 8.3 filename.
                         Defaults to `FAT_OEM_ENCODING` as per FAT spec.
        """
        self.name: bytes = None
        self.encoding = encoding
        self.initialised = False

    @_init_check
    def byte_repr(self):
        """Byte representation of the 8DOT3 name dir entry headers."""
        return self.name

    @_init_check
    def get_unpadded_filename(self) -> str:
        """Retrieve the human readable filename."""
        base = self.name[:8].decode(self.encoding).rstrip()
        ext = self.name[8:11].decode(self.encoding).rstrip()
        sep = "." if len(ext) > 0 else ""

        return sep.join([base, ext])

    @staticmethod
    def __raise_8dot3_nonconformant(name: str):
        raise PyFATException(f"Given directory name "
                             f"{name} is not conform "
                             f"to 8.3 file naming convention.",
                             errno=errno.EINVAL)

    def __set_name(self, name: bytes):
        """Set self.name and verify for correctness."""
        name_str = name.decode(self.encoding)
        if len(name_str) != 11:
            self.__raise_8dot3_nonconformant(name_str)

        self.name = name
        self.initialised = True

    def set_byte_name(self, name: bytes):
        """Set the name as byte input from a directory entry header.

        :param name: `bytes`: Padded (must be 11 bytes) 8dot3 name
        """
        if not isinstance(name, bytes):
            raise TypeError(f"Given parameter must be of type bytes,"
                            f"but got {type(name)} instead.")

        if len(name) != 11:
            raise ValueError("Invalid byte name supplied, must be exactly "
                             "11 bytes long (8+3).")

        if name[0] == 0x0 or name[0] == 0xE5:
            # Empty directory entry
            raise NotAFatEntryException("Given dir entry is invalid and has "
                                        "no valid name.", free_type=name[0])

        if name[0] == 0x05:
            # Translate 0x05 to 0xE5
            name = name.replace(bytes(0x05), bytes(0xE5), 1)

        self.__set_name(name)

    def set_str_name(self, name: str):
        """Set the name as string from user input (i.e. folder creation)."""
        if not isinstance(name, str):
            raise TypeError(f"Given parameter must be of type str,"
                            f"but got {type(name)} instead.")

        if not self.is_8dot3_conform(name):
            self.__raise_8dot3_nonconformant(name)

        name = self._pad_8dot3_name(name)
        self.__set_name(name.encode(self.encoding))

    @_init_check
    def checksum(self) -> int:
        """Calculate checksum of byte string.

        :returns: Checksum as int
        """
        chksum = 0
        for c in self.name:
            chksum = ((chksum >> 1) | (chksum & 1) << 7) + c
            chksum &= 0xFF
        return chksum

    @staticmethod
    def __check_characters(name: str) -> bool:
        """Test if given string contains invalid chars for 8.3 names.

        :param name: `str`: Filename to parse
        :raises: `ValueError` if the given string contains invalid
                 8.3 filename characters.
        """
        vc = EightDotThree.VALID_CHARACTERS
        n = ''.join(filter(lambda x: x in vc, name))
        if n != name:
            raise ValueError(f"Invalid characters in string '{name}', "
                             f"cannot be used as part of an 8.3 "
                             f"conform file name.")

    @staticmethod
    def is_8dot3_conform(entry_name: str):
        """Indicate conformance of given entries name to 8.3 standard.

        :param entry_name: Name of entry to check
        :returns: bool indicating conformance of name to 8.3 standard
        """
        if entry_name != entry_name.upper():
            # Case sensitivity check
            return False

        root, ext = os.path.splitext(entry_name)
        ext = ext[1:]
        if len(root) + len(ext) > 11:
            return False
        elif len(root) > 8 or len(ext) > 3:
            return False

        # Check for valid characters in both filename segments
        for i in [root, ext]:
            try:
                EightDotThree.__check_characters(i)
            except ValueError:
                return False

        return True

    @staticmethod
    def _pad_8dot3_name(name: str):
        """Pad 8DOT3 name to 11 bytes for header operations.

        This is required to pass the correct value to the `FATDirectoryEntry`
        constructor as a DIR_Name.
        """
        root, ext = os.path.splitext(name)
        ext = ext[1:]
        name = root.strip().ljust(8) + ext.strip().ljust(3)
        return name

    @staticmethod
    def make_8dot3_name(dir_name: str,
                        parent_dir_entry) -> str:
        """Generate filename based on 8.3 rules out of a long file name.

        In 8.3 notation we try to use the first 6 characters and
        fill the rest with a tilde, followed by a number (starting
        at 1). If that entry is already given, we increment this
        number and try again until all possibilities are exhausted
        (i.e. A~999999.TXT).

        :param dir_name: Long name of directory entry.
        :param parent_dir_entry: `FATDirectoryEntry`: Dir entry of parent dir.
        :returns: `str`: 8DOT3 compliant filename.
        :raises: PyFATException: If parent dir is not a directory
                                 or all name generation possibilities
                                 are exhausted
        """
        dirs, files, _ = parent_dir_entry.get_entries()
        dir_entries = [e.get_short_name() for e in dirs + files]

        extsep = "."

        def map_chars(char: chr) -> chr:
            """Map 8DOT3 valid characters.

            :param char: `str`: input character
            :returns: `str`: mapped output character
            """
            char = char.upper()
            if char == ' ':
                return ''
            if char not in EightDotThree.VALID_CHARACTERS:
                return '_'
            return char

        try:
            # Shorten to 8 chars; strip invalid characters
            basename = os.path.splitext(dir_name)[0][0:8].strip()
            basename = ''.join(map(map_chars, basename))
        except IndexError:
            basename = ""

        try:
            # Shorten to 3 chars; strip invalid characters
            extname = os.path.splitext(dir_name)[1][1:4].strip()
            extname = ''.join(map(map_chars, extname))
        except IndexError:
            extname = ""

        if len(extname) == 0:
            extsep = ""

        # Loop until suiting name is found
        i = 0
        while len(str(i)) + 1 <= 7:
            if i > 0:
                maxlen = 8 - (1 + len(str(i)))
                basename = f"{basename[0:maxlen]}~{i}"

            short_name = f"{basename}{extsep}{extname}"

            if short_name not in dir_entries:
                return short_name
            i += 1

        raise PyFATException("Cannot generate 8dot3 filename, "
                             "unable to find suiting short file name.",
                             errno=errno.EEXIST)
